@@ -18,7 +18,7 @@
 #include "catalog/pg_type.h"
 #include "mb/pg_wchar.h"
 
-const char *PROGRAM_VERSION	= "1.0.2";
+const char *PROGRAM_VERSION	= "2.0.0";
 const char *PROGRAM_URL		= "http://pgstatsinfo.projects.postgresql.org/";
 const char *PROGRAM_EMAIL	= "pgstatsinfo-general@pgfoundry.org";
 
@@ -30,7 +30,7 @@ const char *PROGRAM_EMAIL	= "pgstatsinfo-general@pgfoundry.org";
 typedef struct Format
 {
 	char			   *name;
-	xmlXPathObjectPtr (*func)(PGresult *rs, xmlNsPtr ns, const char *name);
+	xmlXPathObjectPtr (*func)(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
 } Format;
 
 /* local defs */
@@ -38,19 +38,20 @@ static void pg_query(xmlXPathParserContextPtr ctxt, int nargs);
 static void addNode(xmlNodePtr parent, const char *name, const char *value, Oid type);
 static void toArray(StringInfo str, PGresult *rs);
 
-static xmlXPathObjectPtr createTABLE(PGresult *rs, xmlNsPtr ns, const char *name);
-static xmlXPathObjectPtr createTBODY(PGresult *rs, xmlNsPtr ns, const char *name);
-static xmlXPathObjectPtr createDL(PGresult *rs, xmlNsPtr ns, const char *name);
-static xmlXPathObjectPtr createULOL(PGresult *rs, xmlNsPtr ns, const char *name);
-static xmlXPathObjectPtr createSPAN(PGresult *rs, xmlNsPtr ns, const char *name);
-static xmlXPathObjectPtr createPRE_postgres(PGresult *rs, xmlNsPtr ns, const char *name);
-static xmlXPathObjectPtr createPRE_mysql(PGresult *rs, xmlNsPtr ns, const char *name);
-static xmlXPathObjectPtr createPRE_oracle(PGresult *rs, xmlNsPtr ns, const char *name);
+static xmlXPathObjectPtr createTABLE(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
+static xmlXPathObjectPtr createTBODY(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
+static xmlXPathObjectPtr createDL(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
+static xmlXPathObjectPtr createULOL(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
+static xmlXPathObjectPtr createSPAN(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
+static xmlXPathObjectPtr createPRE_postgres(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
+static xmlXPathObjectPtr createPRE_mysql(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
+static xmlXPathObjectPtr createPRE_oracle(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
 
-static xmlXPathObjectPtr createCIRCLE(PGresult *rs, xmlNsPtr ns, const char *name);
-static xmlXPathObjectPtr createVBAR(PGresult *rs, xmlNsPtr ns, const char *name);
-static xmlXPathObjectPtr createLINE(PGresult *rs, xmlNsPtr ns, const char *name);
-static xmlXPathObjectPtr createRADAR(PGresult *rs, xmlNsPtr ns, const char *name);
+static xmlXPathObjectPtr createCIRCLE(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
+static xmlXPathObjectPtr createVBAR(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
+static xmlXPathObjectPtr createLINE(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
+static xmlXPathObjectPtr createRADAR(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
+static xmlXPathObjectPtr createSELECT(PGresult *rs, xmlNsPtr ns, const char *name, const char *id);
 
 static const Format formats[] =
 {
@@ -67,17 +68,19 @@ static const Format formats[] =
 	{ "vbar", createVBAR },
 	{ "line", createLINE },
 	{ "radar", createRADAR },
+	{ "select", createSELECT },
 };
 
 static char	   *input = NULL;			/* input path */
 static char	   *output = NULL;			/* output path */
 static char	   *stylesheet = NULL;		/* stylesheet path */
-static char	   *httpd_dir = NULL;		/* httpd dir */
+static bool	   *httpd = false;			/* httpd mode */
 static int		httpd_port = 8080;		/* httpd port */
 static char	   *httpd_listen = NULL;	/* httpd listen addresses */
 static char	   *init = NULL;			/* init target directory */
 static char	   *parameters = NULL;		/* arguments from command line */
 static List	   *pg_query_vars = NIL;	/* query variables */
+static char    *config = NULL;          /* configuration file */
 
 /* arguments */
 static struct pgut_option options[] =
@@ -86,10 +89,11 @@ static struct pgut_option options[] =
 	{ 's', 'o', "output", &output },
 	{ 's', 's', "stylesheet", &stylesheet },
 	{ 's', 'a', "args", &parameters },
-	{ 's', 'D', "httpd-dir", &httpd_dir },
+	{ 'b', 'H', "httpd", &httpd },
 	{ 'i', 'P', "httpd-port", &httpd_port },
 	{ 's', 'L', "httpd-listen", &httpd_listen },
 	{ 's', 'i', "init", &init },
+	{ 's', 'c', "config", &config },
 	{ 0 }
 };
 
@@ -104,14 +108,15 @@ pgut_help(bool details)
 		return;
 
 	printf("\nReport options:\n");
-	printf("  -t, --template=TEMPLATE   template path\n");
-	printf("  -o, --output=OUTPUT       output path\n");
-	printf("  -s, --stylesheet=XSLT     stylesheet path\n");
+	printf("  -t, --template=TEMPLATE   template file\n");
+	printf("  -o, --output=OUTPUT       output file\n");
+	printf("  -s, --stylesheet=XSLT     stylesheet file\n");
 	printf("  -a, --args=\"key=val&...\"  template parameters\n");
 	printf("\nHttpd options:\n");
-	printf("  -D, --httpd-dir=DIR       httpd dir\n");
+	printf("  --httpd                   httpd mode\n");
 	printf("  -P, --httpd-port=PORT     port number (default:8080)\n");
 	printf("  -L, --httpd-listen=ADDR   listen address (default:localhost)\n");
+	printf("  -c, --config=CONFIG       configuration file (default:./pg_reporter.conf)\n");
 	printf("\nInit options:\n");
 	printf("  -i, --init=DIR            location to be initialized\n");
 }
@@ -147,10 +152,11 @@ main(int argc, char *argv[])
 		/* --init */
 		return init_main(init, argv[0]);
 	}
-	else if (httpd_dir != NULL)
+	else if (httpd)
 	{
+		getDatabases(config);
 		/* httpd mode */
-		return httpd_main(httpd_dir, httpd_port, httpd_listen);
+		return httpd_main(httpd_port, httpd_listen);
 	}
 	else if (input != NULL)
 	{
@@ -172,6 +178,8 @@ main(int argc, char *argv[])
 		/* cgi mode */
 		const char *document_root;
 
+		getDatabases(config);
+
 		if ((document_root = getenv("DOCUMENT_ROOT")) == NULL)
 		{
 			/* no command line and no environment variables -- maybe misusage */
@@ -179,7 +187,7 @@ main(int argc, char *argv[])
 			return 1;
 		}
 
-		return cgi_main(document_root);
+		return cgi_main();
 	}
 
 	return 0;
@@ -502,15 +510,15 @@ execute_with_vars(const char *query, List *vars)
 static void
 pg_query(xmlXPathParserContextPtr ctxt, int nargs)
 {
-	xmlXPathObjectPtr	args[2];
-	const char		   *values[2];
+	xmlXPathObjectPtr	args[3];
+	const char		   *values[3];
 	PGresult		   *rs;
 	int					i;
 
 	/* first argument is query text. ramains are bind vars. */
 	if (nargs <= 0)
 		return;	/* no result */
-	if (nargs > 2)
+	if (nargs > 3)
 	{
 		elog(WARNING, "too many args for pg:query() : %d passed", nargs);
 		return;	/* no result */
@@ -534,7 +542,6 @@ pg_query(xmlXPathParserContextPtr ctxt, int nargs)
 
 	/* execute query in read-only transaction */
 	command("BEGIN READ ONLY", 0, NULL);
-	command("LOCK TABLE statsrepo.instance IN SHARE MODE", 0, NULL);
 	rs = execute_with_vars(values[0], pg_query_vars);
 	if (PQresultStatus(rs) != PGRES_TUPLES_OK)
 	{
@@ -571,7 +578,7 @@ pg_query(xmlXPathParserContextPtr ctxt, int nargs)
 			if (i >= lengthof(formats))
 				elog(WARNING, "invalid format '%s'", values[1]);
 		}
-		ret = format->func(rs, (ns ? *ns : NULL), format->name);
+		ret = format->func(rs, (ns ? *ns : NULL), format->name, values[2]);
 		if (ret != NULL)
 			valuePush(ctxt, ret);
 	}
@@ -603,7 +610,7 @@ done:
  * </table>
  */
 static xmlXPathObjectPtr
-createTABLE(PGresult *rs, xmlNsPtr ns, const char *name)
+createTABLE(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	xmlNodePtr			table, thead, tbody, tr;
 	int					r, rows;
@@ -651,7 +658,7 @@ createTABLE(PGresult *rs, xmlNsPtr ns, const char *name)
  * </tbody>
  */
 static xmlXPathObjectPtr
-createTBODY(PGresult *rs, xmlNsPtr ns, const char *name)
+createTBODY(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	xmlNodePtr			tbody, tr;
 	int					r, rows;
@@ -683,7 +690,7 @@ createTBODY(PGresult *rs, xmlNsPtr ns, const char *name)
  * </dl>
  */
 static xmlXPathObjectPtr
-createDL(PGresult *rs, xmlNsPtr ns, const char *name)
+createDL(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	xmlNodePtr			dl;
 	int					c, cols;
@@ -707,7 +714,7 @@ createDL(PGresult *rs, xmlNsPtr ns, const char *name)
  * </ul or ol>
  */
 static xmlXPathObjectPtr
-createULOL(PGresult *rs, xmlNsPtr ns, const char *name)
+createULOL(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	xmlNodePtr		ul;
 	int				r, rows;
@@ -725,7 +732,7 @@ createULOL(PGresult *rs, xmlNsPtr ns, const char *name)
  * <span>value, ...</span>
  */
 static xmlXPathObjectPtr
-createSPAN(PGresult *rs, xmlNsPtr ns, const char *name)
+createSPAN(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	StringInfoData	str;
 	xmlNodePtr		span;
@@ -909,25 +916,25 @@ createPRE(PGresult *rs, xmlNsPtr ns, const char *name, char sep, bool border)
  * <pre>value, ...</pre>
  */
 static xmlXPathObjectPtr
-createPRE_postgres(PGresult *rs, xmlNsPtr ns, const char *name)
+createPRE_postgres(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	return createPRE(rs, ns, name, '|', false);
 }
 
 static xmlXPathObjectPtr
-createPRE_mysql(PGresult *rs, xmlNsPtr ns, const char *name)
+createPRE_mysql(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	return createPRE(rs, ns, name, '|', true);
 }
 
 static xmlXPathObjectPtr
-createPRE_oracle(PGresult *rs, xmlNsPtr ns, const char *name)
+createPRE_oracle(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	return createPRE(rs, ns, name, ' ', false);
 }
 
 static xmlXPathObjectPtr
-createCIRCLE(PGresult *rs, xmlNsPtr ns, const char *name)
+createCIRCLE(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	xmlNodePtr		graph;
 	StringInfoData	values;
@@ -943,7 +950,7 @@ createCIRCLE(PGresult *rs, xmlNsPtr ns, const char *name)
 }
 
 static xmlXPathObjectPtr
-createVBAR(PGresult *rs, xmlNsPtr ns, const char *name)
+createVBAR(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	xmlNodePtr		graph;
 	StringInfoData	values;
@@ -997,7 +1004,22 @@ appendVars(StringInfo vars, double vMin, double vMax)
 	 * When the difference is too small, use yMin and yMax from value range.
 	 * Otherwise, expand display range to include zero.
 	 */
-	if (vMax - vMin < Max(Abs(vMin), Abs(vMax)) / 10)
+	if (vMax == vMin)
+	{
+		if (vMax == 0.0)
+		{
+			dy = 0.2;
+			yMax = 1.0;
+			yMin = 0.0;
+		}
+		else
+		{
+			dy = ceilby(vMax, LINE_NUM_Y);
+			yMax = (floor(vMax / dy) + 1) * dy;
+			yMin = 0;
+		}
+	}
+	else if (vMax - vMin < Max(Abs(vMin), Abs(vMax)) / 10)
 	{
 		dy = ceilby(vMax - vMin, LINE_NUM_Y);
 		yMax = (floor(vMax / dy) + 1) * dy;
@@ -1019,7 +1041,7 @@ appendVars(StringInfo vars, double vMin, double vMax)
 	appendStringInfo(vars, ",\nyMax: %f", yMax);
 	appendStringInfo(vars, ",\nyMin: %f", yMin);
 	/* y */
-	if (vMin != vMax)
+	if (yMin != yMax)
 	{
 		int		i;
 
@@ -1193,7 +1215,7 @@ is2D(PGresult *rs)
 }
 
 static xmlXPathObjectPtr
-createLINE(PGresult *rs, xmlNsPtr ns, const char *name)
+createLINE(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	xmlNodePtr		graph;
 	StringInfoData	values;
@@ -1216,7 +1238,7 @@ createLINE(PGresult *rs, xmlNsPtr ns, const char *name)
 }
 
 static xmlXPathObjectPtr
-createRADAR(PGresult *rs, xmlNsPtr ns, const char *name)
+createRADAR(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
 {
 	xmlNodePtr		graph;
 	StringInfoData	values;
@@ -1247,6 +1269,42 @@ createRADAR(PGresult *rs, xmlNsPtr ns, const char *name)
 	return xmlXPathNewNodeSet(graph);
 }
 
+static xmlXPathObjectPtr
+createSELECT(PGresult *rs, xmlNsPtr ns, const char *name, const char *id)
+{
+	xmlNodePtr			select, option;
+	xmlAttrPtr			value, identify, selected;
+	int					r, rows;
+	int					cols;
+	bool				select_attr = false;
+
+	rows = PQntuples(rs);
+	cols = PQnfields(rs);
+
+	select = xmlNewNode(ns, (const xmlChar *) "select");
+	identify = xmlNewProp(select, (const xmlChar *) "name",
+						  (const xmlChar *)id);
+
+	for (r = 0 ; r < rows ; r++ )
+	{
+		option = xmlNewChild(select, NULL, (const xmlChar *) "option",
+							 (const xmlChar *) PQgetvalue(rs, r, 0));
+		value = xmlNewProp(option, (const xmlChar *) "value",
+							 (const xmlChar *) PQgetvalue(rs, r, 0));
+		if (cols == 2 && !select_attr)
+			if (strcmp(PQgetvalue(rs, r, 0), PQgetvalue(rs, r, 1)) == 0)
+			{
+				selected = xmlNewProp(option, (const xmlChar *) "selected",
+									  (const xmlChar *) "selected");
+				select_attr = true;
+			}
+
+	}
+
+	return xmlXPathNewNodeSet(select);
+
+}
+
 static void
 toArray(StringInfo str, PGresult *rs)
 {
